{% if question_list %}
<ul>
{% for question in question_list %}
  <li><a href="{% url 'polls:question-detail' question.pk %}">{{ question.question_text }}</a></li>
{% endfor %}
</ul>
{% else %}
<p>No polls are available.</p>
{% endif %}
<p>
    <a href="{% url 'polls:question-new' %}">Add new question</a>
</p>
