<form action="{% url 'polls:question-detail' question.pk %}" method="post">
    {% csrf_token %}
    {{ form }}
    {{ choice_set }}
    <p>
        <input type="submit" value="Update"></input>
    </p>
</form>
