from .__version__ import (
    __author__,
    __author_email__,
    __description__,
    __version__,
)


__all__ = [
    "__author__",
    "__author_email__",
    "__description__",
    "__version__",
]
