<h1>{{ question.question_text }}</h1>

<ul>
{% for choice in question.choices %}
    <li>{{ choice.choice_text }} -- {{ choice.votes }} vote{{ choice.votes|pluralize }}</li>
{% endfor %}
</ul>

<a href="{% url 'polls:question-detail' question.pk %}">Vote again?</a>
